;;; ================================================================================================
;;; Things to test:
;;; o Direction bits
;;; o CLUT < 256 entries and not starting at entry 0
;;; o Small TGAs such that multiple lines can fit in one DMA packet
;;; o RLE packets that span lines
;;; ================================================================================================

	MAX_PACKETS_LOG2	= 3	; this assures that MAX_PACKETS is a power of 2
	MAX_PACKETS		= (1<<MAX_PACKETS_LOG2)
	PACKET_BUFFER_MASK	= ($FFFF0000 | ((MAX_PACKETS*16)-1))

	RLE_BUFF_SCALARS	= 32
	RLE_BUFF_BYTES		= (RLE_BUFF_SCALARS*4)

	;; TGA header offsets
	TGAHDR_IDSIZE		= 0
	TGAHDR_CMAPTYPE		= 1
	TGAHDR_IMGTYPE		= 2
	TGAHDR_CMAPSTART_LO	= 3
	TGAHDR_CMAPSTART_HI	= 4
	TGAHDR_CMAPLEN_LO	= 5
	TGAHDR_CMAPLEN_HI	= 6
	TGAHDR_CMAPBITS		= 7
	TGAHDR_XSTART_LO	= 8
	TGAHDR_XSTART_HI	= 9
	TGAHDR_YSTART_LO	= 10
	TGAHDR_YSTART_HI	= 11
	TGAHDR_WIDTH_LO		= 12
	TGAHDR_WIDTH_HI		= 13
	TGAHDR_HEIGHT_LO	= 14
	TGAHDR_HEIGHT_HI	= 15
	TGAHDR_BITSPERPIXEL	= 16
	TGAHDR_DESCRIPTOR	= 17

	;; Return codes
	RET_DONE		= 2
	RET_ERROR_CMAPTYPE	= 3
	RET_ERROR_IMGTYPE	= 4
	RET_ERROR_CMAPBITS	= 5
	RET_ERROR_BITSPERPIXEL	= 6
	RET_ERROR_DESCRIPTOR	= 7
	RET_ERROR_MISALIGNED	= 8
	RET_ERROR_UNKNOWN	= 9
	
;; =================================================================================================

	tmp_v0			= v0
	tmp_r0			= r0
	tmp_r1			= r1
	tmp_r2			= r2
	tmp_r3			= r3
	
	TGAPacket_v1		= v1
	TGAPckt_Src_r4		= r4		; source TGA (raw TGA file)
	TGAPckt_Dst_r5		= r5		; destination buffer (bilinear address)
	TGAPckt_Flags_r6	= r6		; DMA flags to use for destination writes
	TGAPckt_Ret_r7		= r7		; value to return once processing is done

	;; For CLUT conversion
	CLUTSrcPtr_r8		= r8		; pointer to RGB CLUT data from TGA
	CLUTDstPtr_r9		= r9		; pointer to final YCrCb CLUT data
	CLUT_AlphaSkip_r10	= r10		; 1 if alpha is present (32bpp), 0 if not (24bpp)

	CLUT_RGB_v3		= v3		; Original source color
	CLUT_R_r12		= r12		; ...
	CLUT_G_r13		= r13		; ...
	CLUT_B_r14		= r14		; ...
	CLUT_k_r15		= r15		; constant necessary for matrix multiplies
	
	ConvToY_v4		= v4		; vectors to hold conversion matricies
	ConvToCr_v5		= v5		; ...
	ConvToCb_v6		= v6		; ...

	CLUT_YCrCb_v7		= v7
	CLUT_Y_r28		= r28		; converted color
	CLUT_Cr_r29		= r29		; ...
	CLUT_Cb_r30		= r30		; ...

	;; For rendering
	xCur_r8			= r8		; current X coordinate
	yCur_r9			= r9		; current Y coordinate
	w_r10			= r10		; TGA width
	h_r11			= r11		; TGA height (decremented as we write)
	hRemaining_r11		= r11		; TGA height (decremented as we write)

	xInc_r12		= r12		; 1 for left-to-right, -1 for right-to-left
	yInc_r13		= r13		; 1 for top-to-bottom, -1 for bottom-to-top
	wRemaining_r14		= r14		; number of pixels left to write on this row
	xStart_r15		= r15		; starting X coordinate for each line

	RLESrcPtr_r16		= r16		; pointer to TGA data in memory
	RLELocalPtr_r17		= r17		; pointer within the RLE buffers locally
	RLEBytesLeft_r18	= r18		; number of bytes left to process in this buffer
	RLEStatus_r19		= r19		; 0=new packet, 1=repeating color, 2=unique pixels
		kRLEStatusNew	= 0
		kRLEStatusUniq	= 1
		kRLEStatusRep	= 2

	RLERepeatCount_r20	= r20		; number of pixels to write for this RLE packet
	PixelPtr_r21		= r21		; address where to store rendered pixels
	PixelsToWrite_r22	= r22		; number of pixels ready to be written out
	PixelLimit_r23		= r23		; number of pixels to wait for before starting DMA
	
	RLEColor1_r24		= r24		; pixel color 1
	RLEColor2_r25		= r25		; pixel color 2
	TotalPixelsLeft_r26	= r26		; total number of pixels left to render
	
;; =================================================================================================
	
	.segment tgad
	.origin 0x20100000

	.align (MAX_PACKETS*16*2)
CommRecvBuffer:		.ds.v	MAX_PACKETS
	
CommRecvRdPtr:		.dc.s	CommRecvBuffer
CommRecvWrPtr:		.dc.s	CommRecvBuffer
CommRecvNumPackets:	.dc.s	0
CommRecvPrevHandler:	.dc.s	0

TGAHeader:		.ds.s	5		; 18 byte header (5 scalars)

	DMAOffFlags	= 0
	DMAOffSysAddr	= 4
	ODMAOffMPEAddr	= 8
	MDMAOffXW	= 8
	MDMAOffYH	= 12
	MDMAOffMPEAddr	= 16
	.align.v
DMACmdRd:		.ds.s	4		; buffer of reading TGA info
	.align.v
DMACmdWr:		.ds.s	5		; buffer for writing pixels to pixmap

RLEBuffer1:		.ds.s	RLE_BUFF_SCALARS ; TGA RLE info to be decoded
RLEBuffer2:		.ds.s	RLE_BUFF_SCALARS ; double buffered to be speedy

RLEPtrRead:		.dc.s	RLEBuffer1	; for double buffering these get swapped every loop
RLEPtrProcess:		.dc.s	RLEBuffer2	; ...

PixelBufferWrite:	.dc.s	RenderedPixels1	; for double buffering the rendered pixels
PixelBufferProcess:	.dc.s	RenderedPixels2	; ...

MaxLinesToWrite:	.ds.s	1		; holds the maximum number of lines that fit in
						; a DMA packet (64 scalars)

CommSenderID:		.ds.s	1		; comm bus ID that sent us the packet
NumProcessed:		.dc.s	0		; DEBUG: number of TGAs processed so far
		
	.include "tgaconv.s"
	
TGAClut:		.ds.s	256		; Color Look-Up Table

	;; Need minimum of 64 scalars to max out DMA.  But if we have 63 and the next RLE byte
	;; indicates a 128 pixel chunk then we need room to store 191.
TGATemp:
RenderedPixels1:	.ds.s	192		; place to store the pixels as they are rendered
RenderedPixels2:	.ds.s	192		; double buffer

_End:			.ds.s	1

;; =================================================================================================
	
	.segment tgac
	.origin 0x20300000

;;; 
;;; Get comm bus receive handler based on what MPE we are
;;;
SetUpCommRecvHandler:
.if (0)
	mv_s	#1,tmp_r0
`hang:
	btst	#0,tmp_r0
	bra	ne,`hang,nop
.endif

	ld_s	configa,tmp_r0			; find out what MPE we are
	nop
	bits	#7,>>#8,tmp_r0			; isolate MPE number
	bra	ne,`notMPE0,nop			; skip next section if not MPE0

	ld_s	intvec1,tmp_r1			; int vector 1 has handler addresses
 	nop
 	sub	#20,tmp_r1			; vector - 20 is MINIcommhook
	mv_s	#CommRecvHandler,r0
	jsr	(tmp_r1),nop
 	st_s	tmp_r0,CommRecvPrevHandler	; save previous comm handler
	bra	WaitingForPacket,nop

`notMPE0:
	st_s	#0x20101000,sp			; set up stack, but only when not MPE0
	st_s	#CommRecvHandler,intvec1	; set interrupt 1 vector for comm bus
	st_s	#1<<4,inten1			; ...
	
	st_s	#$55,intctl			; allow interrupt masks

;;; 
;;; Wait for a comm bus packet (interrupt based)
;;; 
WaitingForPacket:
	ld_s	CommRecvNumPackets,r0		; check if a packet is waiting to be processed
	nop
	cmp	#0,r0
	bra	eq,WaitingForPacket,nop

	;; Hang if processed the given number of images
	ld_s	NumProcessed,tmp_r0
	nop
	add	#1,tmp_r0
	st_s	tmp_r0,NumProcessed
.if (0)
	cmp	#8,tmp_r0
	bra	ne,`nohang,nop
	mv_s	#1,tmp_r0
`hang:
	btst	#0,tmp_r0
	bra	ne,`hang,nop
`nohang:
.endif
;;; 
;;; Read the packet
;;; 
ReadTGAPacket:
	
	st_s	#(1<<3),intctl			; mask interrupts
	
	ld_s	CommRecvNumPackets,r0		; decrement the number of received packets
	nop
	sub	#1,r0
	st_s	r0,CommRecvNumPackets
	
	ld_s	CommRecvRdPtr,r1		; get a pointer to the packet
	nop
	ld_v	(r1),TGAPacket_v1		; load the packet
	add	#16,r1				; increment the packet read pointer
	and	#PACKET_BUFFER_MASK,r1		; mask to wrap around pointer (avoids cmp/bra)
	st_s	r1,CommRecvRdPtr		; store the new pointer

	st_s	#(1<<2),intctl			; allow interrupts again

;;; 
;;; Verify the address
;;; 
VerifySrcAddress:
	
	mv_s	#RET_ERROR_MISALIGNED,tmp_r0	; pre-load error code in case the branch happens
	ftst	#3,TGAPckt_Src_r4		; make sure source TGA address is scalar aligned
	bra	ne,DecodeFailed,nop

	;; Store some info in the DMA packet since it never changes
	st_s	TGAPckt_Flags_r6,DMACmdWr+DMAOffFlags	; MDMA flags
	st_s	TGAPckt_Dst_r5,DMACmdWr+DMAOffSysAddr	; MDMA write address
	
;;; 
;;; Read the TGA header
;;; 
ReadHeader:
	
	st_s	#(5<<16)|(1<<13),DMACmdRd+DMAOffFlags	; header is 18 bytes so read 5 scalars
	st_s	TGAPckt_Src_r4,DMACmdRd+DMAOffSysAddr	; address where to find the TGA
	st_s	#TGAHeader,DMACmdRd+ODMAOffMPEAddr	; local address where to store the header
	st_s	#DMACmdRd,odmacptr			; start the DMA
`wait:
	ld_s	odmactl,tmp_r0
	nop
	ftst	#$1F,tmp_r0			; wait until DMA is complete
	bra	ne,`wait,nop

;;; 
;;; Verify that the TGA can be processed.  Eventually I hope to make this generic enough that
;;; it can process every possible type of TGA.
;;; 
VerifyTGA:
	
	ld_b	TGAHeader+TGAHDR_CMAPTYPE,tmp_r1
	ld_b	TGAHeader+TGAHDR_IMGTYPE,tmp_r2
;	ld_b	TGAHeader+TGAHDR_CMAPBITS,tmp_r3
	nop
	mv_s	#RET_ERROR_CMAPTYPE,tmp_r0
	cmp	#1<<24,tmp_r1			; color map must be present
	bra	ne,DecodeFailed
	mv_s	#RET_ERROR_IMGTYPE,tmp_r0
	cmp	#9<<24,tmp_r2			; must be indexed RLE
	bra	ne,DecodeFailed
;	mv_s	#RET_ERROR_CMAPBITS,tmp_r0
;	cmp	#24<<24,tmp_r3			; color map must be 24 bits
;	bra	ne,DecodeFailed
	
	ld_b	TGAHeader+TGAHDR_BITSPERPIXEL,tmp_r1
;	ld_b	TGAHeader+TGAHDR_DESCRIPTOR,tmp_r2
	nop
	mv_s	#RET_ERROR_BITSPERPIXEL,tmp_r0
	cmp	#8<<24,tmp_r1			; data must be 8-bit indexed
	bra	ne,DecodeFailed
;	mv_s	#RET_ERROR_DESCRIPTOR,tmp_r0
;	btst	#4,tmp_r2			; cannot handle right-to-left drawing
;	bra	ne,DecodeFailed

;;; 
;;; Read the CLUT
;;; 
ReadCLUT:
	
	;; Get the number of scalars to read
	ld_b	TGAHeader+TGAHDR_CMAPBITS,tmp_r0
	nop
	lsr	#(24+3),tmp_r0			; # bits / 8 = bytes per clut entry
	lsr	#2,tmp_r0,CLUT_AlphaSkip_r10	; will have 1 if CLUT is 32bpp
	mv_s	#(256/4),tmp_r1
	mul	tmp_r1,tmp_r0,>>#0,tmp_r0	; bytes * 256 / 4 bytes per scalar = num scalars
	nop
	add	#1,tmp_r0			; add one in case clut is not scalar aligned

	;; Get the starting address for the clut - it need not be scalar aligned
	ld_b	TGAHeader+TGAHDR_IDSIZE,tmp_r1
	add	#18,TGAPckt_Src_r4,tmp_r2	; skip over header
	lsr	#24,tmp_r1
	add	tmp_r1,tmp_r2			; skip over ID
	and	#$FFFFFFFC,tmp_r2		; round down to previous scalar boundary

	;; Get the destination address
	mv_s	#TGATemp,tmp_r1			; initialize the destination address

	;; Now loop through reading in batches of CLUT entries.  Need to break it up due to the
	;; requirement of only 32 scalars per transfer.
`ReadNextBatch:
	st_s	tmp_r2,DMACmdRd+DMAOffSysAddr	; store the address from where to read
	st_s	tmp_r1,DMACmdRd+ODMAOffMPEAddr	; store the destination address
	copy	tmp_r0,tmp_r3			; assume we won't clamp
	cmp	#32,tmp_r0			; can only read 32 scalars at a time on ODMA
	bra	le,`NoClamp,nop
	mv_s	#32,tmp_r3			; clamp to 32 scalars
`NoClamp:
	sub	tmp_r3,tmp_r0			; subtract the number of scalars left to read
	add	tmp_r3,>>#-2,tmp_r2		; increment the source address for next loop
	add	tmp_r3,>>#-2,tmp_r1		; increment the destination address for next loop
	lsl	#16,tmp_r3			; put number of scalars in correct slot
	or	#(1<<13),tmp_r3			; set read bit
	st_s	tmp_r3,DMACmdRd+DMAOffFlags	; store the flags
	st_s	#DMACmdRd,odmacptr		; start the DMA to read the CLUT
`wait1:
	ld_s	odmactl,tmp_r3
	nop
	btst	#4,tmp_r3			; wait until DMA is available
	bra	ne,`wait1,nop

	cmp	#0,tmp_r0			; see if we're done reading all scalars
	bra	ne,`ReadNextBatch,nop		; jump back if not done reading

	;; Wait for all DMAs to complete
`wait2:
	ld_s	odmactl,tmp_r0
	nop
	ftst	#$1F,tmp_r0			; wait until DMA is complete
	bra	ne,`wait2,nop

	;; Now we have the RGB CLUT.  Convert to YCrCb and 32bpp.
	mv_s	#TGATemp,CLUTSrcPtr_r8		; get pointer to CLUT we just read
	ld_b	TGAHeader+TGAHDR_IDSIZE,tmp_r0
	nop
	lsr	#24,tmp_r0			; put ID size in bottom byte
	add	#18,tmp_r0			; add 18 byte header
	and	#3,tmp_r0			; get byte offset into source buffer
	add	tmp_r0,CLUTSrcPtr_r8		; add offset to source pointer

	mv_s	#TGAClut,CLUTDstPtr_r9		; get destination pointer for converted colors
	ld_b	TGAHeader+TGAHDR_CMAPSTART_LO,tmp_r0 ; get the first CLUT entry in the TGA
	ld_b	TGAHeader+TGAHDR_CMAPSTART_HI,tmp_r1 ; get upper byte
	lsr	#(24-2),tmp_r0			; get offset to start of first CLUT entry
	or	tmp_r1,>>#(16-2),tmp_r0		; assemble the 2-byte offset
	add	tmp_r0,CLUTDstPtr_r9		; add offset to base address

	ld_b	TGAHeader+TGAHDR_CMAPLEN_LO,tmp_r0 ; get the number of CLUT entries in the TGA
	ld_b	TGAHeader+TGAHDR_CMAPLEN_HI,tmp_r1 ; get upper byte
	lsr	#24,tmp_r0			; put it in [7:0]
	lsr	#16,tmp_r1			; put it in [15:8]
	or	tmp_r1,tmp_r0			; assemble 2-byte length
	st_s	tmp_r0,rc0			; use the handy-dandy RC counter
	
	ld_v	ConvToY,ConvToY_v4		; get conversion matrices
	ld_v	ConvToCr,ConvToCr_v5		; ...
	ld_v	ConvToCb,ConvToCb_v6		; ...
	nop
	mv_s	#(1<<24),CLUT_k_r15		; need a constant for the fourth component

	st_s	#(4<<20),linpixctl		; use type 4 pixels

`NextCLUTEntry:
	ld_b	(CLUTSrcPtr_r8),CLUT_B_r14	; read the source R value
	add	#1,CLUTSrcPtr_r8
	ld_b	(CLUTSrcPtr_r8),CLUT_G_r13	; read the source G value
	add	#1,CLUTSrcPtr_r8
	ld_b	(CLUTSrcPtr_r8),CLUT_R_r12	; read the source B value
	add	#1,CLUTSrcPtr_r8
	add	CLUT_AlphaSkip_r10,CLUTSrcPtr_r8 ; skip alpha if present

	lsr	#2,CLUT_R_r12			; put it in the format we need
	lsr	#2,CLUT_G_r13
	lsr	#2,CLUT_B_r14
	dotp	CLUT_RGB_v3,ConvToY_v4,>>#24,CLUT_Y_r28		; do the matrix multiply to...
	dotp	CLUT_RGB_v3,ConvToCr_v5,>>#24,CLUT_Cr_r29	; ...convert from RGB to YCrCb
	dotp	CLUT_RGB_v3,ConvToCb_v6,>>#24,CLUT_Cb_r30
	nop
	add	#1<<21,CLUT_Y_r28		; round up
	add	#1<<21,CLUT_Cr_r29
	add	#1<<21,CLUT_Cb_r30
	
	st_p	CLUT_YCrCb_v7,(CLUTDstPtr_r9)	; store converted color
	add	#4,CLUTDstPtr_r9		; advance to next destination entry
	
	dec	rc0				; decrement number of entries to process
	bra	c0ne,`NextCLUTEntry,nop		; branch back if more entries left to process

;;; 
;;; Done reading and converting the CLUT.  Now set up parameters for TGA decode.
;;; 
RenderSetup:
	
	mv_s	#0,xStart_r15			; starting coordinates
	mv_s	#0,yCur_r9			; ...
	ld_b	TGAHeader+TGAHDR_WIDTH_LO,w_r10  ; get TGA width
	ld_b	TGAHeader+TGAHDR_WIDTH_HI,tmp_r0 ; get upper byte
	lsr	#24,w_r10
	lsr	#16,tmp_r0
	or	tmp_r0,w_r10
	ld_b	TGAHeader+TGAHDR_HEIGHT_LO,h_r11  ; get TGA height
	ld_b	TGAHeader+TGAHDR_HEIGHT_HI,tmp_r0 ; get upper byte
	lsr	#24,h_r11
	lsr	#16,tmp_r0
	or	tmp_r0,h_r11
	copy	w_r10,TotalPixelsLeft_r26	; compute total number of pixels
	mul	h_r11,TotalPixelsLeft_r26,>>#0,TotalPixelsLeft_r26

	;; Compute the maximum number of lines that can fit in one DMA packet (64 scalars)
	mv_s	#0,tmp_r0			; assume we can't fit a full line
	mv_s	#64,tmp_r1			; figure out how many lines fit in 64 scalars
`CheckMaxLines:
	cmp	tmp_r1,w_r10			; see if another line will fit in the remaining size
	bra	le,`CheckMaxLines		; branch back if a line will fit
	sub	w_r10,tmp_r1			; subtract the width of one line
	add	#1,tmp_r0			; increment the line count
	sub	#1,tmp_r0			; went one too far when we fell through the branch
	st_s	tmp_r0,MaxLinesToWrite		; store the line count
	
	;; Get incrementors for walking through the TGA
	ld_b	TGAHeader+TGAHDR_DESCRIPTOR,tmp_r0
	nop
	mv_s	#1,xInc_r12			; assume left-to-right
	btst	#4+24,tmp_r0
	bra	eq,LeftToRight,nop		; if bit 4 is clear then it is left-to-right
	mv_s	#-1,xInc_r12			; bit was set, it's right-to-left
	copy	w_r10,xStart_r15		; if right-to-left then x coord start at max
	ld_s	DMACmdWr+DMAOffFlags,tmp_r1	; load DMA flags to set backwards flag A
	sub	#1,xStart_r15			; point to last pixel on the line
	bset	#0,tmp_r1			; set backwards flag A
	st_s	tmp_r1,DMACmdWr+DMAOffFlags	; store the new flags
LeftToRight:
	mv_s	#1,yInc_r13			; assume top-to-bottom
	btst	#5+24,tmp_r0
	bra	ne,TopToBottom,nop		; if bit 5 is set then it is top-to-bottom
	mv_s	#-1,yInc_r13			; bit was clear, it's bottom-to-top
	copy	h_r11,yCur_r9			; if bottom-to-top then y coord start at max
	ld_s	DMACmdWr+DMAOffFlags,tmp_r1	; load DMA flags to set backwards flag B
	sub	#1,yCur_r9			; point to last line of the image
	bset	#9,tmp_r1			; set backwards flag B
	st_s	tmp_r1,DMACmdWr+DMAOffFlags	; store the new flags
TopToBottom:
	copy	xStart_r15,xCur_r8

	;; Get the starting address for the RLE pixels
	add	#18,TGAPckt_Src_r4,RLESrcPtr_r16 ; pointer to RLE pixels in memory
	ld_b	TGAHeader+TGAHDR_CMAPLEN_LO,tmp_r0
	ld_b	TGAHeader+TGAHDR_CMAPLEN_HI,tmp_r1
	lsr	#24,tmp_r0			; get number of CLUT entries, lower byte
	lsr	#16,tmp_r1			; get upper byte
	or	tmp_r1,tmp_r0
	ld_b	TGAHeader+TGAHDR_CMAPBITS,tmp_r1
	ld_b	TGAHeader+TGAHDR_IDSIZE,tmp_r2
	lsr	#(24+3),tmp_r1			; get number of bytes per CLUT entry
	mul	tmp_r0,tmp_r1,>>#0,tmp_r1	; get CLUT size in bytes
	lsr	#24,tmp_r2			; put ID size in lower byte
	add	tmp_r2,RLESrcPtr_r16		; add the ID size to the offset
	add	tmp_r1,RLESrcPtr_r16		; add the CLUT size to the offset
	and	#3,RLESrcPtr_r16,tmp_r0		; number of bytes to skip if not scalar aligned
	and	#~3,RLESrcPtr_r16		; round down to scalar boundary

	mv_s	#kRLEStatusNew,RLEStatus_r19	; need to get first RLE byte to determine type
	mv_s	#0,PixelsToWrite_r22		; no pixels in the buffer yet
	mv_s	#64,PixelLimit_r23		; max DMA transfer is 64 scalars
	cmp	PixelLimit_r23,TotalPixelsLeft_r26
	bra	ge,`GotLimit,nop		; clamp limit to number of pixels left to render
	copy	TotalPixelsLeft_r26,PixelLimit_r23
`GotLimit:
	ld_s	PixelBufferProcess,PixelPtr_r21	; address where to store rendered pixels

;;; 
;;; Preload the first set of RLE pixels
;;; 
PreloadFirstBuffer:
	
	ld_s	RLEPtrRead,RLELocalPtr_r17	; get address where to store RLE info locally
	st_s	#(RLE_BUFF_SCALARS<<16)|(1<<13),DMACmdRd+DMAOffFlags
	st_s	RLESrcPtr_r16,DMACmdRd+DMAOffSysAddr ; address from where to get the RLE data
	st_s	RLELocalPtr_r17,DMACmdRd+ODMAOffMPEAddr	; local buffer pointer
	st_s	#DMACmdRd,odmacptr		; start the DMA
	
	add	tmp_r0,RLELocalPtr_r17		; offset based on the skipped bytes
	mv_s	#RLE_BUFF_BYTES,RLEBytesLeft_r18 ; number of bytes to decode
	sub	tmp_r0,RLEBytesLeft_r18		; subtract off the skipped bytes
	st_s	RLEBytesLeft_r18,rc0		; count of bytes to process

;;; 
;;; Load next set of RLE pixels in the background
;;; 
GetNextRLEData:
		
	add	#RLE_BUFF_BYTES,RLESrcPtr_r16	; increment our source pointer for the next load
	
	ld_s	RLEPtrRead,tmp_r0		; swap the buffers
	ld_s	RLEPtrProcess,tmp_r1
	st_s	tmp_r0,RLEPtrProcess
	st_s	tmp_r1,RLEPtrRead

	;; Wait for previous RLE data to be done reading
`wait:
	ld_s	odmactl,tmp_r0
	nop
	ftst	#$1F,tmp_r0			; wait until DMA is complete
	bra	ne,`wait,nop
	
	ld_s	RLEPtrRead,tmp_r0		; get address where to store next bunch of RLE bytes
	st_s	RLESrcPtr_r16,DMACmdRd+DMAOffSysAddr ; address from where to get the RLE data
	st_s	tmp_r0,DMACmdRd+ODMAOffMPEAddr	; local buffer pointer
	st_s	#DMACmdRd,odmacptr		; start the DMA to preload the next set
	
;;; 
;;; Process the RLE bytes
;;; 

CheckRLEStatus:
	;; Check if we have enough pixels to write
	cmp	PixelLimit_r23,PixelsToWrite_r22
	bra	ge,DMAPixels,nop

	;; Check what type of RLE packet we're processing
	cmp	#kRLEStatusRep,RLEStatus_r19
	bra	eq,ProcessRepeatingColor,nop
	
	cmp	#kRLEStatusUniq,RLEStatus_r19
	bra	eq,ProcessUniquePixels,nop

	;; Don't have status yet, grab the next RLE byte to get it.
GetNextRLEType:
	ld_b	(RLELocalPtr_r17),RLERepeatCount_r20
	nop
	lsr	#24,RLERepeatCount_r20
	lsr	#7,RLERepeatCount_r20,RLEStatus_r19 ; 0=unique pixels, 1=repeating color
	add	#1,RLEStatus_r19		; 1=unique pixels, 2=repeating color
	and	#$7F,RLERepeatCount_r20		; isolate the count for this packet
	add	#1,RLERepeatCount_r20		; count is actually value+1
	add	#1,RLELocalPtr_r17		; advance to next RLE byte
	dec	rc0				; decrement number of available RLE bytes
	bra	c0ne,CheckRLEStatus,nop		; jump back if more RLE bytes available
	bra	RLEBufferDone,nop		; otherwise grab the next RLE buffer

	;; =====================================================================================
	;; Write out a single repeating color.  Write vectors when possible.  This will help for 
	;; long strings of pixels, but hurt for short strings.  See TGAVectorAnalysis.txt for 
	;; details.
	;; =====================================================================================
ProcessRepeatingColor:
	ld_b	(RLELocalPtr_r17),tmp_r0	; get CLUT entry for the color
	add	#1,RLELocalPtr_r17		; advance to next RLE byte
	lsr	#24-2,tmp_r0			; convert to an offset within the CLUT
	add	#TGAClut,tmp_r0			; now have the address for the color
	ld_s	(tmp_r0),RLEColor1_r24		; load the color
	dec	rc0				; decrement RLE byte count

	;; Increment the number of pixels written
	add	RLERepeatCount_r20,PixelsToWrite_r22

	;; For really short strings of pixels (5 or less) it's usually faster to not use vectors.
	;; See TGAVectorAnalysis.txt
	cmp	#5,RLERepeatCount_r20
	bra	gt,`RepeatSingle,nop

	;; Eliminate the cmp/bra to the vector writing routine so we save ticks.
`RepeatSingleOnly:
	sub	#1,RLERepeatCount_r20		; decrement the repeat count
	bra	ne,`RepeatSingleOnly		; jump back if not done
	st_s	RLEColor1_r24,(PixelPtr_r21)	; otherwise store them individually
	add	#4,PixelPtr_r21			; increment address where to store the next pixel
	bra	`RepeatDone,nop			; quit if done

	;; Start out writing single pixels.  Jump to writing vectors at the first opportunity.
`RepeatSingle:
	ftst	#$F,PixelPtr_r21		; check if pointer is on vector boundary
	bra	eq,`WriteVectors,nop		; if so we can speedify things by writing vectors
	sub	#1,RLERepeatCount_r20		; decrement the repeat count
	bra	ne,`RepeatSingle		; jump back if not done
	st_s	RLEColor1_r24,(PixelPtr_r21)	; otherwise store them individually
	add	#4,PixelPtr_r21			; increment address where to store the next pixel
	mv_s	#RET_ERROR_UNKNOWN,tmp_r0	; load the error code
	bra	DecodeFailed,nop		; should never get here

	;; Blast vectors of the same color.
`WriteVectors:
{
	subm	tmp_r3,tmp_r3			; need constant zero
	lsl	#2,RLERepeatCount_r20		; convert from # of pixels to # of scalars
}
`RepeatVector:
{
	sub	#16,RLERepeatCount_r20		; decrement the pixel count (scalar count now)
	addm	RLEColor1_r24,tmp_r3,tmp_r0	; copy color to vector[0]
	mv_s	RLEColor1_r24,tmp_r1		; copy color to vector[1]
}
{
	bra	gt,`RepeatVector		; jump back if not done
	addm	RLEColor1_r24,tmp_r3,tmp_r2	; copy color to vector[2]
	mv_s	RLEColor1_r24,tmp_r3		; copy color to vector[3]
}
	st_v	v0,(PixelPtr_r21)		; store a vector of colors (might write extra)
{
	add	#16,PixelPtr_r21		; increment the address
	subm	tmp_r3,tmp_r3			; reset this back to zero
}
	;; Fall through when done writing vectors

	;; Now correct if we overshot when writing vectors
`VectorCorrect:
	add	RLERepeatCount_r20,PixelPtr_r21	; back up pixel write address (value is negative)
	
`RepeatDone:
	mv_s	#0,RLEStatus_r19		; reset the status
	bra	c0ne,CheckRLEStatus,nop		; if more bytes are available then go parse them
	bra	RLEBufferDone,nop		; otherwise grab the next RLE buffer

	
ProcessUniquePixels:
`Repeat:
	ld_b	(RLELocalPtr_r17),tmp_r0	; get CLUT entry for the color
	add	#1,RLELocalPtr_r17		; advance to next RLE byte
	lsr	#(24-2),tmp_r0			; convert to an offset within the CLUT
	add	#TGAClut,tmp_r0			; add the base CLUT address
	ld_s	(tmp_r0),RLEColor1_r24		; read the color
	sub	#1,RLERepeatCount_r20		; decrement the repeat count
{
	st_s	RLEColor1_r24,(PixelPtr_r21)	; store pixel
	bra	eq,`RepeatDone			; jump out if done
}
	add	#4,PixelPtr_r21			; increment address where to store the next pixel
{
	add	#1,PixelsToWrite_r22		; just wrote one pixel
	dec	rc0				; decrement count of RLE bytes
}
	bra	c0ne,`Repeat,nop		; jump back if we still have RLE bytes left
	bra	RLEBufferDone,nop		; oops, ran out of RLE bytes, better read some more
	
`RepeatDone:
	mv_s	#0,RLEStatus_r19		; reset the status
	bra	c0ne,CheckRLEStatus,nop
	bra	RLEBufferDone,nop
	
;;; 
;;; Done with this bunch of RLE bytes.
;;; 
RLEBufferDone:
	
	ld_s	RLEPtrRead,RLELocalPtr_r17	; will process what was just loaded
	nop
	mv_s	#RLE_BUFF_BYTES,RLEBytesLeft_r18 ; reset bytes left to decode
	st_s	RLEBytesLeft_r18,rc0		; count of bytes to process
	bra	GetNextRLEData,nop


;;; 
;;; DMA pixels to pixmap
;;;
DMAPixels:
	ld_s	PixelBufferProcess,PixelPtr_r21	; swap the buffers
	ld_s	PixelBufferWrite,tmp_r1
	st_s	PixelPtr_r21,PixelBufferWrite
	st_s	tmp_r1,PixelBufferProcess

	ld_s	MaxLinesToWrite,tmp_r2		; maximum number of lines we can write in a packet
DMAMorePixels:
	;; Check if we can try to write entire lines.  The current X pointer must be at the
	;; beginning of a line and the number of pixels available must be at least the width of 
	;; the image.  Also, we computed earlier how many lines will fit in a 64-scalar DMA
	;; packet.  If that's less than two then don't bother.
	cmp	xStart_r15,xCur_r8
	bra	ne,WritePartialLines,nop	; branch if current X pointer is not at the start
	cmp	w_r10,PixelsToWrite_r22
	bra	lt,WritePartialLines,nop	; branch if number of pixels is less than the width
	cmp	#2,tmp_r2
	bra	lt,WritePartialLines,nop	; branch if we can't write at least two lines

	;; =====================================================================================
	;; When writing full lines, we are still restricted to 64 scalars per DMA transfer.  If
	;; the image width is greater than 64 pixels than we can't write full lines at all and
	;; shouldn't even get here.  If the width is less than or equal to 64 pixels then we
	;; must make sure the number of lines we write doesn't total more than 64 pixels.  We
	;; precomputed the max number of lines earlier and use that as a comparison here.
	;; =====================================================================================
WriteFullLines:
	;; Get X and W components
	copy	xStart_r15,tmp_r0		; construct the DMA X/W component
	or	w_r10,>>#-16,tmp_r0
	st_s	tmp_r0,DMACmdWr+MDMAOffXW
	
`WriteMoreLines:
	st_s	PixelPtr_r21,DMACmdWr+MDMAOffMPEAddr ; store the address where pixels will be taken
	eor	tmp_r1,tmp_r1			; zero out the number of lines to write
	copy	yCur_r9,tmp_r3			; we clobber this later so make a copy of it
`AddLine:
	add	yInc_r13,yCur_r9		; increment/decrement the Y coordinate for next time
	sub	#1,hRemaining_r11		; decrement the number of lines remaining
	add	#1,tmp_r1			; increment number of lines to write
	add	w_r10,>>#-2,PixelPtr_r21	; increment the address from where to pull pixels
	sub	w_r10,PixelsToWrite_r22		; decrement the number of pixels available
	sub	w_r10,TotalPixelsLeft_r26	; decrement the total number of pixels to be written
	cmp	tmp_r1,tmp_r2			; see if we've hit the maximum number of lines
	bra	eq,`DoneAddingLines,nop
	cmp	w_r10,PixelsToWrite_r22		; see if we can write at least one more line
	bra	ge,`AddLine,nop

`DoneAddingLines:
	lsl	#16,tmp_r1			; put H in top 16 bits
	or	tmp_r3,tmp_r1			; fold in the starting Y coordinate
	st_s	tmp_r1,DMACmdWr+MDMAOffYH	; store the Y/W component
	st_s	#DMACmdWr,mdmacptr		; start the DMA to write pixels

	;; Wait for MDMA to be free
`wait:
	ld_s	mdmactl,tmp_r0
	nop
	btst	#4,tmp_r0			; wait until DMA is available
	bra	ne,`wait,nop

	;; Check if we've written all the lines
	cmp	#0,hRemaining_r11
	bra	le,DecodeSuccessful,nop
	
	;; In the case when we are restricted to the number of lines we can write at one time,
	;; check to see if there are more full lines to write.
	cmp	w_r10,PixelsToWrite_r22		; see if we can write at least one more line
	bra	ge,`WriteMoreLines,nop		; jump back to write another block

	;; Check if there's more pixels to write at all.
	cmp	#0,PixelsToWrite_r22
	bra	eq,DMADone,nop
	;; Fall through if there are more pixels to write

	;; =====================================================================================
	;; When writing partial lines we can write at most 64 pixels due to MDMA restrictions.
	;; If the number of pixels remaining on the line is less than this then it will be
	;; clamped.  It will also be clamped to the number of pixels available to be written.
	;; 
	;; Once we reach a new line we have the option to jump up to the routine to write blocks
	;; of lines.
	;; =====================================================================================
WritePartialLines:
	;; Figure out how many pixels we can write on this line
	st_s	PixelPtr_r21,DMACmdWr+MDMAOffMPEAddr ; store the address where pixels will be taken
	mv_s	#64,tmp_r2			; assume max for now
	sub	xStart_r15,xCur_r8,tmp_r1	; compute the number of pixels remaining on the line
	abs	tmp_r1				; make it a positive number when going right-to-left
	sub	tmp_r1,w_r10,tmp_r1		; subtract this from the width
	cmp	tmp_r2,tmp_r1			; check which is lower
	bra	gt,`NoClampToWidth,nop		; if the remaining pixels > 64 then don't clamp
	copy	tmp_r1,tmp_r2			; otherwise clamp to the size of the line
`NoClampToWidth:
	cmp	tmp_r2,PixelsToWrite_r22	; now clamp to the number of pixels we actually have
	bra	gt,`NoClampToNumAvail,nop	; don't clamp if there are enough pixels available
	copy	PixelsToWrite_r22,tmp_r2	; otherwise clamp to the number of pixels we have
`NoClampToNumAvail:
	
	;; Get X and W components
	lsl	#16,tmp_r2			; put number of pixels in upper 16 bits
	or	xCur_r8,tmp_r2			; construct the DMA X/W component
	st_s	tmp_r2,DMACmdWr+MDMAOffXW	; store X/W component

	;; Get Y and H components
	or	#1<<16,yCur_r9,tmp_r1		; construct the DMA Y/H component (H always = 1)
	st_s	tmp_r1,DMACmdWr+MDMAOffYH	; store Y/H component
	st_s	#DMACmdWr,mdmacptr		; start the DMA to write pixels

	;; Wait for MDMA to be free
`wait:
	ld_s	mdmactl,tmp_r0
	nop
	btst	#4,tmp_r0			; wait until DMA is available
	bra	ne,`wait,nop

	;; Now increment all the things that need incrementing (pointers, coordinates, etc)
	lsr	#16,tmp_r2			; get number of pixels written
	copy	tmp_r2,tmp_r1			; make a copy for the multiply
	mul	xInc_r12,tmp_r1,>>#0,tmp_r1	; factor in left-to-right or right-to-left
	nop
	add	tmp_r1,xCur_r8			; increment/decrement the X coordinate for next time
	add	tmp_r2,>>#-2,PixelPtr_r21	; increment the address from where to pull pixels
	sub	tmp_r2,PixelsToWrite_r22	; decrement the number of pixels available
	sub	tmp_r2,TotalPixelsLeft_r26	; decrement the total number of pixels to be written
	sub	xStart_r15,xCur_r8,tmp_r1	; compute the number of pixels remaining on the line
	abs	tmp_r1				; make it positive number when going right-to-left
	cmp	tmp_r1,w_r10			; see if we're done with this line
;	bra	ne,WritePartialLines,nop	; line not done - go back and do another partial DMA
	bra	ne,`ContinueLine,nop		; line not done - go back and do another partial DMA

	;; Get here when a partial DMA finishes a line
	add	yInc_r13,yCur_r9		; increment/decrement the Y coordinate for next time
	copy	xStart_r15,xCur_r8		; start over with the X coordinate
	sub	#1,hRemaining_r11		; decrement the number of lines remaining
	bra	eq,DecodeSuccessful,nop
`ContinueLine:
	cmp	#0,PixelsToWrite_r22		; see if there are pixels left to process
	bra	ne,DMAMorePixels,nop		; head back up to the start, maybe using full lines

DMADone:
	ld_s	PixelBufferProcess,PixelPtr_r21
	cmp	PixelLimit_r23,TotalPixelsLeft_r26
	bra	ge,`GotLimit,nop		; clamp limit to number of TGA pixels
	copy	TotalPixelsLeft_r26,PixelLimit_r23
`GotLimit:
	bra	CheckRLEStatus,nop		; continue processing pixels
	
;;; 
;;; Decode complete
;;;
	
DecodeSuccessful:
	mv_s	#RET_DONE,tmp_r0		; processing completed successfully
	
DecodeFailed:
	copy	TGAPckt_Ret_r7,tmp_r1		; return the value we were given

	ld_s	configa,tmp_r3			; find out what MPE we are
	ld_s	CommSenderID,tmp_r2		; get the comm bus ID that sent us this packet
	bits	#7,>>#8,tmp_r3			; isolate MPE number
	bra	ne,`notMPE0,nop			; skip next section if not MPE0
	
	ld_s	intvec1,tmp_r1			; int vector 1 has handler addresses
 	nop
 	sub	#16,tmp_r1			; vector - 16 is MINIcommsend
	jsr	(tmp_r1)
	copy	tmp_r2,r4
	bra	WaitingForPacket,nop
	
`notMPE0:
	
`CommWait:
	ld_s	commctl,tmp_r3			; poll for transmit data empty
	nop
	btst	#15,tmp_r3			; ready to send?
	bra	ne,`CommWait,nop		; no -- keep waiting

`Resend:
	st_s	tmp_r2,commctl			; send to the MPE that started it
	st_v	v0,commxmit			; start the transfer
	
`CommXmitWait:
	;; Wait for the data to go
	ld_s	commctl,tmp_r3
	nop
	btst	#14,tmp_r3			; check for transmit failure
{	bra	ne,`Resend,nop			; failed -- retry
	btst	#15,tmp_r3			; check for transmit buffer empty
}
	bra	eq,WaitingForPacket,nop		; all done - go back and get the next TGA packet
	bra	`CommXmitWait,nop

	
;;; 
;;; Local comm bus receive handler
;;;
	
CommRecvHandler:
	push	v0
	ld_s	CommRecvNumPackets,r0		; get the number of packets pending
	ld_s	CommRecvWrPtr,r1		; pointer where to put more packets
	cmp	#MAX_PACKETS,r0			; if buffer is full then quit
	bra	eq,`NoMoreRoom,nop
	push	v1
	ld_s	commctl,r2			; get the sender's comm bus ID
	ld_v	commrecv,v1			; read the packet
	bits	#7,>>#16,r2			; isolate the sender's comm bus ID
	st_s	r2,CommSenderID			; store it for later
	add	#1,r0				; increment number of packets
	st_s	r0,CommRecvNumPackets		; store the new packet count
	st_v	v1,(r1)				; store the packet data
	add	#16,r1				; increment the write pointer
	and	#PACKET_BUFFER_MASK,r1		; mask to wrap around pointer (avoids cmp/bra)
	st_s	r1,CommRecvWrPtr		; store the new pointer
	st_s	#1<<4,intclr			; clear the interrupt
	pop	v1
	rti	(rzi1)
	pop	v0
	nop
`NoMoreRoom:
	rti	(rzi1)				; no more room for packets so don't read yet
	pop	v0
	st_s	#(1<<3),intctl			; mask interrupts, will be cleared once there's room
